import os, subprocess, multiprocessing, sys
import time
from os import listdir
from os.path import isfile, join
import json
import sys
sys.path.append('.')
import expr, data, tptp
from stratified import is_guarded, is_pvd, is_stratified, args, is_epr, is_2var
from expr import Visitor, Var, Fun
import shutil

# global settings
terminationTool = "./ttt2"

def get_problem_features():
  f = open("features.json", "r")
  return json.loads(f.read())

def literals(f):
  if f.is_literal():
    return [f]
  else:
    assert(isinstance(f, expr.Binop))
    return literals(f.left) + literals(f.right)
  
def fold_left(func, init, seq):
  acc = init
  for i in seq:
    acc = func(acc, i)
  return acc

def forall(pred, seq):
  return fold_left(lambda v, l: v and pred(l), True, seq)

def exists(pred, seq):
  return fold_left(lambda v, l: v or pred(l), False, seq)

def find(pred, seq):
  return fold_left(lambda v, l: l if pred(l) else v, None, seq)

def cross(lss):
  return fold_left(lambda rss, ls: [ rs + [l] for rs in rss for l in ls], [[]], lss)

def all_ground(ls):
  return forall(lambda l: l.is_ground(), ls)

def covered(ls, is_pos):
  poss = [l for l in ls if l.is_positive()]
  negs = [l for l in ls if not l.is_positive()]
  (cover, to_cover) = (negs, poss) if is_pos else (poss, negs)

  def covered(lp):
    return exists(lambda ln: lp.vars().issubset(ln.vars()), cover)

  return forall(lambda lp: lp.is_ground() or covered(lp), to_cover)

def all_all_pos_all_ground(cs):
  non_gnd = [ls for ls in cs if not all_ground(ls)]
  vars_pos_in_neg = forall(lambda c: covered(c, True), non_gnd)
  return vars_pos_in_neg

def all_all_neg_all_ground(cs):
  non_gnd = [ls for ls in cs if not all_ground(ls)]
  vars_neg_in_pos = forall(lambda c: covered(c, False), non_gnd)
  return vars_neg_in_pos

def atom(l):
  return l if isinstance(l, expr.Fun) or l.is_positive() else l.arg

def is_strat(cs):
  return is_stratified(cs)

### create TRSs
def rules_for_clause(ls, is_pos, covered):
  poss = [l for l in ls if l.is_positive()]
  negs = [l for l in ls if not l.is_positive()]
  (cover, to_cover) = (negs, poss) if is_pos else (poss, negs)
  to_cover = [l for l in to_cover if not l.is_ground()] 

  def covering(lp):
    return [ln for ln in cover if lp.vars().issubset(ln.vars())]

  trs = []
  if covered:
    if not(forall(lambda l: covering(l) != [], to_cover)):
      return None
    trs = [[(l,lp) for l in covering(lp)] for lp in to_cover]
  else:
    trs = [[(l,lp) for l in cover] for lp in to_cover]
  return trs

def run_ttt2(f):
  bashCommand = "timeout 10 " + terminationTool + " " + f
  process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
  out, err = process.communicate()
  out = str(out)
  if err:
    print(err)
  if "MAYBE" in out:
    print("MAYBE" + f)
  technique = ""
  if "LPO" in out:
    technique = "LPO"
  elif "KBO" in out:
    technique = "KBO"
  elif "Matrix" in out:
    technique = "Matrix"
  return "YES" in out, technique

def is_permutative(l,r):
  def all_vars(ts):
    return forall(lambda v: isinstance(v, expr.Var), ts)
  var_args = all_vars(l.args) and all_vars(r.args)
  return l.name == r.name and var_args and l.vars() == r.vars()

def check_trs(p, i, trs, nonground = False):
  if not trs: # empty
    return (True, "trivial"), False
  trs = list(set(trs))
  vars = set([v for (l, r) in trs for v in set(l.vars()).union(set(r.vars()))])
  var_str = fold_left(lambda s, v: s + " " + v, "", vars)
  trs_str = "(VAR " + var_str + ")\n"
  trs_str = trs_str + "(RULES \n"
  for (l,r) in trs:
    l = atom(l)
    r = atom(r)
    increasing = l in r.subterms()
    op = "->=" if is_permutative(l,r) or increasing else "->"
    trs_str = trs_str + l.toString() + " " + op + " " + r.toString() + "\n"
  trs_str = trs_str + ")\n"
  #print("  TRS %d" % len(trs))
  filename = "trss/" + p + ".trs"
  rfile = open(filename, "w")
  rfile.write(trs_str)
  rfile.close()
  return run_ttt2(filename), "->=" in trs_str

def get_trss(p, lss, is_pos, covered):
  rss_all = [rules_for_clause(ls, is_pos, covered) for ls in lss]
  if None in rss_all:
    return []
  
  rss = [rs for ls in lss for rsl in rss_all for rs in rsl if rs != []]
  cross_cnt = fold_left(lambda p, l: p * l, 1, [len(rs) for rs in rss ])
  #print(cross_cnt)
  if cross_cnt > 100000: # too large
    trss = [[rs[0] for rs in rss]] # just one TRS
  else:
    trss = cross([rs for rs in rss])
  return trss
  
def mk_gen(syms):
  rs = []
  gen = expr.Fun("gen", [])
  for (f, a) in syms:
    rs.append((gen, expr.Fun(f, [gen for i in range(0,a)])))
  return rs

def check_trss(p, lss, is_pos):
  trss = get_trss(p, lss, is_pos, covered=True)
  proven = False
  technique = ""
  for (i, trs) in enumerate(trss[:50]):
    (proven, technique), _ = check_trs(p, i, trs)
    if proven:
      break
  sys.stdout.flush()
  return proven, technique

# given clauses cs, check whether all-positive or all-negative ground
def check_ground_fixed(p, cs, flipped = False):
  fs = [c.formula for c in cs]
  lss = [literals(c) for c in fs]
  proven = False
  technique = ""
  if all_all_pos_all_ground(lss):
    proven, technique = check_trss(p, lss, True)
  if not proven and all_all_neg_all_ground(lss):
    proven, technique = check_trss(p, lss, False)
  
  restrained = (all_all_pos_all_ground(lss), all_all_neg_all_ground(lss))
  return (restrained, proven, technique, is_guarded(lss), is_pvd(lss), is_strat(fs))

# flip predicate p in f
def flip_formula(p, f):
  def flip(f):
    if isinstance(f, expr.Binop):
      return expr.Binop(f.op, flip(f.left), flip(f.right))
    elif isinstance(f, expr.Unop):
      if isinstance(f.arg, expr.Pred):
        if f.arg.name == p:
          return f.arg 
        else:
          return f
      else:
        print("unexpected for CNF")
        assert(False)
    elif isinstance(f, expr.Pred):
      if f.name == p:
        return expr.Unop("~", f) 
      else:
        return f
    else:
      print("unexpected for CNF")
      assert(False)
  
  fflip = flip(f.formula)
  return expr.Cnf(f.name, f.kind, fflip)

def check_ground(p, formulas):
  res = check_ground_fixed(p, formulas)
  preds = [ atom(l).name for c in formulas for l in literals(c.formula) ]
  if res[1] or len(preds) > 50: # success
    return res, ""
  
  for pred in list(set(preds)):
    formulas_flipped = [ flip_formula(pred,f) for f in formulas]
    res_p_flipped = check_ground_fixed(p, formulas_flipped, flipped = True)
    if res_p_flipped[1]: # success
      return res_p_flipped, "flipped " + pred
  return res, ""

def check(cnfdir, p, tptp_properties, verbose = False):
  not_applicable = ""
  MAX_SIZE = 500
  if tptp_properties["vampire_axiom_clauses"] > MAX_SIZE:
    not_applicable = "too large"
  elif tptp_properties["vampire_equality_atoms"] > 0:
    not_applicable = "equality"
  #print(p)
  res = {"error": not_applicable, "success": False}
  if not not_applicable:
    filename = cnfdir + "/" + p + ".p"
    content = data.get_problem(filename)
    if not content or "negated_conjecture, ($false))." in content:
      res = {"error": "empty", "success": False}
    elif len([l for l in content.splitlines() if "cnf(" in l]) > MAX_SIZE:
      res = {"error": "too large", "success": False}
    else:
      #print(p + ": " + filename)
      formulas = tptp.parse_problem(content, verbose)
      ((bndp, bndn), success, technique, guarded, pvd, strat), msg = check_ground(p, formulas)
      res["pos_restrained"] = bndp
      res["neg_restrained"] = bndn
      res["restrained"] = success
      res["guarded"] = guarded
      res["stratified"] = strat
      res["pvd"] = pvd
      res["technique"] = technique
      res["flipped"] = msg
      res["epr"] = is_epr([f.formula for f in formulas])
      res["ground"] = tptp_properties["num_variables"] == 0
      res["monadic"] = tptp_properties["vampire_max_pred_arity"] <= 1
      res["2var"] = is_2var([f.formula for f in formulas])
    
      res["success"] = True
      status = tptp_properties["status"]
      res["status"] = "SAT" if status == "Satisfiable" else \
        "UNSAT" if status == "Unsatisfiable" else \
        "THM" if status == "Theorem" else \
        "CSAT" if status == "CounterSatisfiable" else "UNK"

  return res

if __name__ == "__main__":
  cnfdir = str(sys.argv[1])
  features = get_problem_features()

  # create directory where TRSs are stored
  if not os.path.exists('trss'):
    os.makedirs('trss')

  fs = {}
  for f in features.keys():
    if not ("=" in f):
      fs[f] = check(cnfdir, f, features[f])
  
  usable = [c for c in fs if fs[c]["success"]]
  print("%d excluded because of equality" % len([f for f in fs if "equality" == fs[f]["error"]]))
  print("%d excluded to prevent timeouts" % len([f for f in fs if "too large" == fs[f]["error"]]))
  print("%d considered:" % len([f for f in usable]))
  print(" %d positively restrained" % len([f for f in usable if fs[f]["pos_restrained"]]))
  print(" %d negatively restrained" % len([f for f in usable if fs[f]["neg_restrained"]]))
  print(" %d restrained" % len([f for f in usable if fs[f]["restrained"]]))
  print(" %d PVD" % len([f for f in usable if fs[f]["pvd"]]))
  print(" %d EPR" % len([f for f in usable if fs[f]["epr"]]))
  print(" %d guarded" % len([f for f in usable if fs[f]["guarded"]]))
  print(" %d monadic" % len([f for f in usable if fs[f]["monadic"]]))
  print(" %d 2-variable" % len([f for f in usable if fs[f]["2var"]]))
  print(" %d stratified" % len([f for f in usable if fs[f]["stratified"]]))
  print("<table>")
  print("<tr><th class=\"thead\">problem</th><th class=\"thead\">status</th><th class=\"thead\">rating</th>")
  print("<th class=\"thead\">restrained</th><th>TRS</th><th class=\"thead\">properties</th><th></th></tr>")
  for u in usable:
    dom = u[0:3]
    href = "http://www.tptp.org/cgi-bin/SeeTPTP?Category=Problems&Domain=" + dom + "&File=" + u + ".p"
    print("<tr>")
    print("<td class=\"system\"><a href=\"%s\">%s</a></td>" % (href, u))
    print("<td>%s</td>" % fs[u]["status"])
    print("<td>%s</td>" % features[u]["rating"])
    # restrained
    bnd = "positive" if fs[u]["pos_restrained"] else "negative"
    trsfilename = "trss/" + u + ".trs"
    print("<td>%s</td>" % (bnd))
    print("<td><a href=\"%s\">TRS</a></td>" % (trsfilename))
    # other properties
    ground = "ground" if features[u]["num_variables"] == 0 else ""
    mon = "monadic" if features[u]["vampire_max_pred_arity"] <= 1 else ""
    two_var = "2var" if features[u]["vampire_max_variables_in_clause"] <= 2 and features[u]["vampire_max_fun_arity"] <= 0 else ""
    guarded = "guarded" if fs[u]["guarded"] else ""
    pvd = "PVD" if fs[u]["pvd"] else ""
    epr = "EPR" if fs[u]["epr"] else ""
    pres = "positively restrained" if fs[u]["pos_restrained"] else ""
    nres = "negatively restrained" if fs[u]["neg_restrained"] else ""
    strat = "stratified" if fs[u]["stratified"] else ""
    none = "<!--none-->" if not(epr or mon or two_var or guarded or ground or pvd or strat or pres or nres) else ""
    #print("%s: %d (%s) (%.2f)  %s %d %s" % (u, bnd, fs[u]["technique"], features[u]["rating"], features[u]["source"], features[u]["vampire_axiom_clauses"], fs[u]["status"]))
    props = [epr, mon, two_var, guarded, ground, pvd, strat, pres, nres, none]
    props = [p for p in props if p]
    ps = fold_left(lambda s, p: s + (", " if s else "") + p, "", props)
    print("<td>%s</td>" % ps)
    sggs = "sggsruns/" + u + ".txt"
    print("<td>%s</td>" % fs[u]["flipped"])
    print("</tr>")
  #r = sum([features[u]["rating"] for u in usable]) / 117.0
  #print("average rating %.2f" % r)
  print("</table>")

  # 